import java.awt.* ;
import java.util.* ;

/***************************************************************/
/* Dmonstration des polygone comme liste circulaire de points */
/* Usage: java Poly, construire un polygone (bouton Close) qui */
/*   doit tre convexe et bien orient, puis ajouter           */
/*   des points.                                               */
/***************************************************************/

class Poly {
  private Point val ; private Poly next ;

  private Poly (Point val, Poly next) { this.val = val ; this.next = next ; }


  /* Lire le polygone contenu dans un PointReader */
  static Poly readPoly(PointReader in) {
    Point pt = in.read() ;
    if (pt == null) return null ;

    Poly r = new Poly (pt, readPoints(in)) ;
    nappend(r,r) ; // Pour boucler la liste, voir exercice 1.4
    return r ;
  }

  static private Poly readPoints(PointReader in) {
    Point pt = in.read() ;
    if (pt == null) {
      return null ;
    } else {
      return new Poly(pt, readPoints(in)) ;
    }
  }

  static Poly nappend(Poly xs, Poly ys) {
    if (xs == null) return ys ;
    Poly last = xs ;
    for (  ; last.next != null ; last = last.next )
      ; // idiome : boucle qui ne fait rien
    last.next = ys ;
    return xs ;
  }

  /* Position du point q relativement au ct orient p */
  static int getSide(Point q, Poly p) {
    Point p1 = p.val ;
    Point p2 = p.next.val ;
    return
      (q.y - p1.y) * (p2.x - p1.x) - (q.x - p1.x) * (p2.y - p1.y) ;
  }

  /* Test de convexit et d'orientation d'un prolygone p */
  static boolean isConvex(Poly p) {
    Poly q = p ;
    do {
      Poly r = q.next.next ;
      do {
        if (getSide(r.val, q) > 0) return false ;
        r = r.next ;
      } while (r != q) ;
      q = q.next ;
    } while (q != p) ;
    return true ;
  }


  /*********************************************/
  /* Calcul incrmental de l'enveloppe convexe */
  /*********************************************/

  static Poly getLeft(Poly p, Point pt) {
    Poly q = p ;
    do {
      if (getSide(pt, q) < 0) return q ;
      q = q.next ;
    } while (q != p) ;
    return null ;
  }


  static Poly getRight(Poly p, Point pt) {
    Poly q = p ;
    do {
      if (getSide(pt, q) > 0) return q ;
      q = q.next ;
    } while (q != p) ;
    return null ;
  }

  static Poly extend(Poly e, Point q) {
    Poly oneRight = getRight(e, q) ;
    if (oneRight == null) return e ; // Point intrieur

    Poly firstLeft = getLeft(oneRight, q) ;
    Poly firstRight = getRight(firstLeft, q) ;
    Poly r = new Poly(q, firstLeft) ;
    firstRight.next = r ;
    return r ;
  }

  /* Demo : points entrs  la souris */
  static void demoIncrement() {
    PolyShow in = new PolyShow() ;
    for ( ; ; ) {
      Poly p = readPoly(in) ;
      in.show(p) ;
      if (isConvex(p)) {
        System.out.println("Polygone convexe et bien orient") ;
        for ( ; ; ) {
          Point pt = in.read() ;
          if (pt == null) break ;
          p = extend(p, pt) ;
          in.show(p) ;
        }
      } else {
        System.out.println("Le polygone n'est pas convexe ou est mal orient") ;

        while (in.read() != null)
          ;
      }
    }
  }


  /*************************************/
  /* Demos, points lus dans un fichier */
  /*************************************/

  static void demoReader(PointReader in, Controler ctrl) {
    Point pt1 = in.read() ; ctrl.show(pt1) ;
    Point pt2 = in.read() ; ctrl.show(pt2) ;
    Poly  p = new Poly (pt1, new Poly (pt2, null)) ;    
    nappend(p,p) ;

    for ( ; ; ) {
      ctrl.show(p) ;
      Point pt = in.read() ;
      if (pt == null) break ;
      ctrl.show(p, pt) ;
      ctrl.show(pt) ;
      p = extend(p, pt) ;
    }
  }

  /* Sans trier */
  static void demoFile(String name) {
    Controler ctrl = new PolyShow(false).ctrl ;
    PointReader in = new PointFileReader (name) ;
    demoReader(in, ctrl) ;
  }
 

  // Pour organiser la communication avec PolyShow et Controler
  private static Poly copy (Poly p, Poly fst) {
    if (p == fst) {
      return null ;
    } else {
      return new Poly (p.val, copy(p.next, fst)) ;
    }
  }

  static Poly copy(Poly p) {
    if (p == null) return null ;
    Poly r = new Poly (p.val, copy(p.next,p)) ;
    nappend(r,r) ;
    return r ;
  }


  PolyIterator getIterator() { return getIterator(this) ; }

  static PolyIterator getIterator(final Poly p) {
    return
      new PolyIterator () {
          private Poly now = p ;
          public Point next() {
            Point r = now.val ;
            now = now.next ;
            return r ;
          }

          public int getSide(Point q) {
            return Poly.getSide(q,now) ;
          }
        } ;
  }

  /* Calcul de l'enveloppe convexe en O(n log n) */
  private Poly prev ; // Liste doublement chane

  private Poly(Point val, Poly next, Poly prev) {
    this.val = val ; this.next = next ; this.prev = prev ;
  }

  static Poly getLeftNext(Poly e, Point q) {
    Poly p = e ;
    do {
      if (getSide(q, p) <= 0) return p ;
      p = p.next ;
    } while (p != e) ;
    return e ;
  }

  static int getSidePrev(Point q, Poly p) {
    Point p1 = p.val ;
    Point p2 = p.prev.val ;
    return
      (q.y - p1.y) * (p2.x - p1.x) - (q.x - p1.x) * (p2.y - p1.y) ;
  }

  static Poly getRightPrev(Poly e, Point q) {
    Poly p = e ;
    do {
      if (getSidePrev(q, p) >= 0) return p ;
      p = p.prev ;
    } while (p != e) ;
    return e ;
  }

  static Poly extend2(Poly e, Point q) {
    Poly ei = getRightPrev(e, q) ;
    Poly ej = getLeftNext(e, q) ;
    System.err.println("ei=" + ei.val + ", ej=" + ej.val) ;
    Poly r = new Poly(q, ej, ei) ;
    if (ei != ej) {
      ei.next = r ; ej.prev = r ;
    } else {
      // Cas particulier d'enfer : tous les pts sont aligns!
      Poly lst = e.prev ;
      r.prev = lst ;
      lst.next = r ;
    }
    return r ;
  }

  /* En triant */
  static void demoSorted(String name) {
    Controler ctrl = new PolyShow(false).ctrl ;
    PointReader in = new PointFileReader (name) ;
    ArrayList<Point> allPoints = new ArrayList<Point> () ;
    Point pt = null ;

    while ((pt = in.read()) != null) {
      allPoints.add(pt) ; ctrl.show(pt) ;
    }
    Collections.sort
      (allPoints,
       new Comparator<Point> () {
        public int compare(Point p1, Point p2) {
          if (p1.x < p2.x) return -1 ;
          else if (p1.x > p2.x) return 1 ;
          else if (p1.y < p2.y) return 1 ;
          else if (p1.y > p2.y) return -1 ;
          else return 0 ;
        }           
      });
    final Iterator<Point> iterator = allPoints.iterator () ;
    demoReader2
      (new PointReader () {
          public Point read () {
            if (iterator.hasNext()) {
              return iterator.next() ;
            } else {
              return null ;
            }
          }
        },
       ctrl) ;
  }

  static void demoReader2(PointReader in, Controler ctrl) {
    Point pt1 = in.read() ; ctrl.show(pt1) ;
    Point pt2 = in.read() ; ctrl.show(pt2) ;
    Poly p1 = new Poly(pt1, null, null) ;
    Poly p2 = new Poly(pt2, p1, p1) ;
    p1.next = p1.prev = p2 ;
    Poly  p = p2 ;
    for ( ; ; ) {
      ctrl.show(p) ;
      System.err.println(p) ;
      Point pt = in.read() ;
      if (pt == null) break ;
      ctrl.show(p, pt) ;
      p = extend2(p, pt) ;
    }
  }

  public String toString() {
    StringBuilder r = new StringBuilder () ;
    Poly p = this ;
    do {
      r.append(p.val + "\n") ;
      p = p.next ;
    } while (p != this) ;
    return r.toString() ;
  }

  public static void main(String [] arg) {
    if (arg.length == 0) {
      demoIncrement() ;
    } else if (arg.length == 1) {
      demoFile(arg[0]) ;
    } else if (arg.length == 2 && arg[0].equals("-s")) {
      demoSorted(arg[1]) ;
    }
  }
}

